/* QNT - buffer.cpp - File buffer class. */


#define WIN32_LEAN_AND_MEAN
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "buffer.h"
#include "config.h"
#include "display.h"
#include "keyboard.h"


char *Buffer::enterstr[4] = { "\n", "\r\n", "\r", "\n\r" };
int Buffer::enterlen[4] = { 1, 2, 1, 2 };


int Buffer::CharList::count( int *corr )
{
    int i = 0;
    if( corr )
        *corr = 0;
    for( CharList *ptr = this; ptr->tab != -1; ptr = ptr->next )
    {
        i++;
        if( corr && ptr->tab == -2 )
            (*corr)++;
    }
    return i;
}


void Buffer::CharList::put( char *b )
{
    for( CharList *ptr = this; ptr->tab != -1; ptr = ptr->next )
    {
        *b = ptr->ch;
        b++;
    }
}


void Buffer::CharList::putrev( char *b )
{
    for( CharList *ptr = this; ptr->tab != -1; ptr = ptr->next )
    {
        *b = ptr->ch;
        b--;
    }
}


Buffer::Buffer()
{
    build();
    rows = 1;
    ent = 0;
    filename = new char[1];
    filename[0] = 0;
    autosavename = 0;
    exitstatus = FALSE;
}


Buffer::~Buffer()
{
    if( truemod )
    {
        display.message( "Save changes? (Y/n)" );
        if( keyboard.yesnoP() )
            save();
    }
    brk();
    delete filename;
    if( autosavename )
    {
        remove( autosavename );
        delete autosavename;
    }
}


void Buffer::build()
{
    startlines = new LineList;
    endlines = new LineList;
    startchars = new CharList;
    endchars = new CharList;
    row = col = drow = dcol = arow = acol = 0;
    tabsize = 8;
    invhilite = invscreen = TRUE;
    prehilite = changed = modified = truemod = FALSE;
}


void Buffer::brk()
{
    delete startlines;
    delete endlines;
    delete startchars;
    delete endchars;
}


void Buffer::postmap( BOOL extsel )
{
    if( changed )
    {
        invscreen = modified = truemod = TRUE;
        changed = FALSE;
    }
    if( extsel )
        invhilite = TRUE;
    else
    {
        arow = row;
        acol = col;
        if( prehilite )
            invhilite = TRUE;
    }
    prehilite = arow != row || acol != col;
}


BOOL Buffer::literal( char c )
{
    if( tabsize == 0 || c != '\t' )
    {
        startchars = new CharList( c, startchars );
        col++;
    }
    else
    {
        int newcol = tabsize * ((col + tabsize) / tabsize);
        startchars = new CharList( c, char( newcol - col ), startchars );
        col = newcol;
    }
    return TRUE;
}


void Buffer::exitline( int &vissize, int &size, char *&buf )
{
    int startsize = startchars->count();
    int corr;
    int endsize = endchars->count( &corr );
    size = startsize + endsize;
    vissize = size - corr;
    buf = new char[size];
    startchars->putrev( buf + startsize - 1 );
    endchars->put( buf + startsize );
    delete startchars;
    delete endchars;
}


void Buffer::enterline( int vissize, int size, char *buf, int prefcol )
{
    startchars = new CharList;
    endchars = new CharList;
    char *ptr = buf;
    for( col = 0; (col < prefcol) && (ptr < buf + vissize); ptr++ )
    {
        if( tabsize == 0 || *ptr != '\t' )
        {
            startchars = new CharList( *ptr, startchars );
            col++;
        }
        else
        {
            int newcol = tabsize * ((col + tabsize) / tabsize);
            startchars = new CharList( '\t', char( newcol - col ), startchars );
            col = newcol;
        }
    }
    for( char *ptr2 = buf + size - 1; ptr2 >= buf + vissize; ptr2-- )
        endchars = new CharList( *ptr2, -2, endchars );
    for( ; ptr2 >= ptr; ptr2-- )
        endchars = new CharList( *ptr2, endchars );
}


char *Buffer::seekcol( char *buf, int cl )
{
    for( int i = 0; i < cl; buf++ )
    {
        if( tabsize == 0 || *buf != '\t' )
            i++;
        else
            i = tabsize * ((i + tabsize) / tabsize);
    }
    return buf;
}


BOOL Buffer::seekto( int prefrow, int prefcol )
{
    if( prefrow >= 0 && prefrow < rows )
    {
        int vissize;
        int size;
        char *buf;
        exitline( vissize, size, buf );
        if( prefrow != row )
        {
            if( prefrow < row )
                while( prefrow < row )
                {
                    endlines = new LineList( vissize, size, buf, endlines );
                    vissize = startlines->getvislen();
                    size = startlines->getlen();
                    buf = startlines->cdrbuf();
                    startlines = startlines->cdr();
                    row--;
                }
            else
                while( prefrow > row )
                {
                    startlines = new LineList( vissize, size, buf, startlines );
                    vissize = endlines->getvislen();
                    size = endlines->getlen();
                    buf = endlines->cdrbuf();
                    endlines = endlines->cdr();
                    row++;
                }
        }
        enterline( vissize, size, buf, prefcol );
        return TRUE;
    }
    else
        return FALSE;
}


void Buffer::prefloc( int prefrow, int prefcol )
{
    prefrow += drow;
    prefcol += dcol;
}
