/* QNT - redraw.cpp - File buffer class / screen redraw. */


#define WIN32_LEAN_AND_MEAN
#include <string.h>
#include <wtypes.h>
#include "buffer.h"
#include "config.h"
#include "display.h"


void Buffer::lineimage( int x, char *buf, LineList *ll )
{
    int len = ll->vislen;
    char *str = ll->str;
    int i = 0;
    while( i < dcol )
    {
        if( len )
        {
            if( tabsize == 0 || *str != '\t' )
            {
                len--;
                str++;
            }
            else
            {
                if( (i + 1) % tabsize == 0 )
                {
                    len--;
                    str++;
                }
            }
            i++;
        }
        else
        {
            i = dcol;
            break;
        }
    }
    while( i < dcol + x )
    {
        if( len )
        {
            if( tabsize == 0 || *str != '\t' )
            {
                len--;
                *buf++ = *str++;
            }
            else
            {
                if( (i + 1) % tabsize == 0 )
                {
                    len--;
                    str++;
                }
                *buf++ = ' ';
            }
        }
        else
            *buf++ = ' ';
        i++;
    }
}


void Buffer::scrimage( int y, int x, char *buf )
{
    char *ptr = buf + (row - drow) * x;
    CharList *cl = endchars;
    for( int i = col - dcol; i < x; i++ )
    {
        if( cl->tab >= 0 )
        {
            if( tabsize == 0 || cl->ch != '\t' )
            {
                ptr[i] = cl->ch;
                cl = cl->next;
            }
            else
            {
                ptr[i] = ' ';
                if( (dcol + i + 1) % tabsize == 0 )
                    cl = cl->next;
            }
        }
        else
            ptr[i] = ' ';
    }
    cl = startchars;
    for( i = col - dcol - 1; i >= 0; i-- )
    {
        if( tabsize == 0 || cl->ch != '\t' )
        {
            ptr[i] = cl->ch;
            cl = cl->next;
        }
        else
        {
            ptr[i] = ' ';
            if( (dcol + i + cl->tab) % tabsize == 0 )
                cl = cl->next;
        }
    }
    LineList *ll = endlines;
    char *scr = ptr + x;
    for( i = row - drow + 1; i < y; i++, scr += x )
    {
        if( ll )
        {
            lineimage( x, scr, ll );
            ll = ll->next;
        }
        else
            memset( scr, ' ', x );
    }
    ll = startlines;
    for( i = row - drow - 1, scr = ptr - x; i >= 0; i--, scr -= x )
    {
        lineimage( x, scr, ll );
        ll = ll->next;
    }
}


void Buffer::refresh()
{
    if( row < drow )
    {
        drow = row;
        invscreen = TRUE;
    }
    else
    {
        int h = display.getheight();
        if( row >= drow + h )
        {
            drow = row - h + 1;
            invscreen = TRUE;
        }
    }
    if( col < dcol )
    {
        dcol = col;
        invscreen = TRUE;
    }
    else
    {
        int w = display.getwidth();
        if( col >= dcol + w )
        {
            dcol = col - w + 1;
            invscreen = TRUE;
        }
    }
    if( invscreen )
    {
        invhilite = TRUE;
        display.redraw( this, Buffer::scrimage );
        invscreen = FALSE;
    }
    if( invhilite )
    {
        if( arow < row || (arow == row && acol < col) )
            display.hilite( SHORT( arow - drow ), SHORT( acol - dcol ),
                SHORT( row - drow ), SHORT( col - dcol ) );
        else
            display.hilite( SHORT( row - drow ), SHORT( col - dcol ),
                SHORT( arow - drow ), SHORT( acol - dcol ) );
        invhilite = FALSE;
    }
    char *ptr = Config::genstatusline( truemod, row, col, filename );
    display.drawstatus( ptr, SHORT( row - drow ), SHORT( col - dcol ) );
    delete [] ptr;
}
